@@ -46,6 +46,7 @@ class DocxTemplate(object):
         self.docx = None
         self.is_rendered = False
         self.is_saved = False
+        self.allow_missing_pics = False
 
     def init_docx(self, reload: bool = True):
         if not self.docx or (self.is_rendered and reload):
@@ -194,7 +195,8 @@ class DocxTemplate(object):
             src_xml = re.sub(pat, r"\1 \2", src_xml, flags=re.DOTALL)
 
         # add vMerge
-        # use {% vm %} to make this table cell and its copies be vertically merged within a {% for %}
+        # use {% vm %} to make this table cell and its copies
+        # be vertically merged within a {% for %}
         def v_merge_tc(m):
             def v_merge(m1):
                 return (
@@ -311,7 +313,7 @@ class DocxTemplate(object):
                 line_number = max(exc.lineno - 4, 0)
                 exc.docx_context = map(
                     lambda x: re.sub(r"<[^>]+>", "", x),
-                    src_xml.splitlines()[line_number : (line_number + 7)],
+                    src_xml.splitlines()[line_number: (line_number + 7)],  # fmt: skip
                 )
 
             raise exc
@@ -370,7 +372,7 @@ class DocxTemplate(object):
                         else part.blob
                     )
                     xml = self.render_xml_part(xml, part, context, jinja_env)
-                    part._blob = xml
+                    part._blob = xml.encode("utf-8")
 
     def resolve_listing(self, xml):
 
@@ -797,10 +799,13 @@ class DocxTemplate(object):
             if rel.reltype in (REL_TYPE.HEADER, REL_TYPE.FOOTER):
                 self._replace_docx_part_pics(rel.target_part, replaced_pics)
 
-        # make sure all template images defined by user were replaced
-        for img_id, replaced in replaced_pics.items():
-            if not replaced:
-                raise ValueError("Picture %s not found in the docx template" % img_id)
+        if not self.allow_missing_pics:
+            # make sure all template images defined by user were replaced
+            for img_id, replaced in replaced_pics.items():
+                if not replaced:
+                    raise ValueError(
+                        "Picture %s not found in the docx template" % img_id
+                    )
 
     def get_pic_map(self):
         return self.pic_map
@@ -860,7 +865,8 @@ class DocxTemplate(object):
                         replaced_pics[img_id] = True
                         break
 
-            # FIXME: figure out what exceptions are thrown here and catch more specific exceptions
+            # FIXME: figure out what exceptions are thrown here
+            # and catch more specific exceptions
             except Exception:
                 continue
 
@@ -881,20 +887,36 @@ class DocxTemplate(object):
         self.is_saved = True
 
     def get_undeclared_template_variables(
-        self, jinja_env: Optional[Environment] = None
+        self,
+        jinja_env: Optional[Environment] = None,
+        context: Optional[Dict[str, Any]] = None,
     ) -> Set[str]:
-        self.init_docx(reload=False)
-        xml = self.get_xml()
+        # Create a temporary document to analyze the template without affecting the current state
+        temp_doc = Document(self.template_file)
+
+        # Get XML from the temporary document
+        xml = self.xml_to_string(temp_doc._element.body)
         xml = self.patch_xml(xml)
+
+        # Add headers and footers
         for uri in [self.HEADER_URI, self.FOOTER_URI]:
-            for relKey, part in self.get_headers_footers(uri):
-                _xml = self.get_part_xml(part)
-                xml += self.patch_xml(_xml)
+            for relKey, val in temp_doc._part.rels.items():
+                if (val.reltype == uri) and (val.target_part.blob):
+                    _xml = self.xml_to_string(parse_xml(val.target_part.blob))
+                    xml += self.patch_xml(_xml)
+
         if jinja_env:
             env = jinja_env
         else:
             env = Environment()
-        parse_content = env.parse(xml)
-        return meta.find_undeclared_variables(parse_content)
 
-    undeclared_template_variables = property(get_undeclared_template_variables)
+        parse_content = env.parse(xml)
+        all_variables = meta.find_undeclared_variables(parse_content)
+
+        # If context is provided, return only variables that are not in the context
+        if context is not None:
+            provided_variables = set(context.keys())
+            return all_variables - provided_variables
+
+        # If no context provided, return all variables (original behavior)
+        return all_variables
